/*++

Copyright (c) Sydus.

Module:

    exceptions.h

Authors:

    Sharvil Nanavati (sharvil) 2005-01-08

--*/

#pragma once

#include <cerrno>
#include <string>
#include <stdexcept>

/* Exception hierarchy:

std::runtime_error
	sydus_error
		invalid_argument
		system_error
		resource_error
			not_found_error
			malformed_error
			constraint_error
			duplicate_error
		permission_error
		io_error
			read_error
			write_error
			network_error
			disconnected_error
*/

struct sydus_error : public std::runtime_error
{
	uint32 iErrNo;
	sydus_error() : runtime_error(""), iErrNo(errno) {}
	sydus_error(const char* msg) : runtime_error(msg), iErrNo(errno) {}
	sydus_error(const std::string& msg) : runtime_error(msg), iErrNo(errno) {}
	virtual ~sydus_error() throw() {}
};

struct null_pointer : public sydus_error
{
	null_pointer(const std::string & error) : sydus_error(error) {}
	virtual ~null_pointer() throw() {}
};

struct invalid_argument : public sydus_error
{
	invalid_argument(const std::string & error) : sydus_error(error) {}
	virtual ~invalid_argument() throw() {}
};

struct system_error : public sydus_error
{
	system_error(const std::string & error) : sydus_error(error) {}
	virtual ~system_error() throw() {}
};

struct resource_error : public sydus_error
{
	resource_error(const std::string & error) : sydus_error(error) {}
	virtual ~resource_error() throw() {}
};

struct not_found_error : public resource_error
{
	not_found_error(const std::string & resourceName) : resource_error(resourceName + " not found.") {}
	virtual ~not_found_error() throw() {}
};

struct malformed_error : public resource_error
{
	malformed_error(const std::string & resourceName, const std::string & error) : resource_error(resourceName + " malformed: " + error) {}
	virtual ~malformed_error() throw() {}
};

struct constraint_error : public resource_error
{
	constraint_error(const std::string & resourceName, const std::string & error) : resource_error(resourceName + " constraint failed: " + error) {}
	virtual ~constraint_error() throw() {}
};

struct duplicate_error : public resource_error
{
	duplicate_error(const std::string & resourceName, const std::string & error) : resource_error("duplicate " + resourceName + ": " + error) {}
	virtual ~duplicate_error() throw() {}
};

struct permission_error : public sydus_error
{
	permission_error(const std::string & error) : sydus_error(error) {}
	virtual ~permission_error() throw() {}
};

struct io_error : public sydus_error
{
	uint32 iSize;
	io_error(uint32 size) : sydus_error(""), iSize(size) {}
    io_error(const std::string& error, uint32 size = 0): sydus_error(error + "("+error+")"), iSize(size) {}
	virtual ~io_error() throw() {}
    typedef sydus_error super;
};

struct write_error : public io_error
{
	write_error(uint32 size) : io_error(size) {}
	virtual ~write_error() throw() {}
};

struct read_error : public io_error
{
	read_error(uint32 size) : io_error(size) {}
    read_error(const std::string& error, uint32 size = 0): io_error(error, size) {}
	virtual ~read_error() throw() {}
};

struct network_error : public io_error
{
	network_error(const std::string & error) : io_error(error) {}
	virtual ~network_error() throw() {}
};

struct disconnected_error : public io_error
{
	disconnected_error(const std::string& error) : io_error(error) {}
	virtual ~disconnected_error() throw() {}
};
