#include "base.h"
#include "ZoneList.h"
#include "SSNetLib.h"
#include "TimeLib.h"
#include "LogLib.h"
#include "Config.h"

#include <sstream>

ZoneList::ZoneList(const Config & config)
	: iLogger(Logger::Instance()),
	  iConfig(config)
{
	iZoneQueryTimer.Start();
}

ZoneList::~ZoneList()
{
}

void ZoneList::AddSource(const InetAddress & address)
{
	for(Sources::iterator i = iSources.begin(); i != iSources.end(); ++i)
		if(*i == address)
			return;

	iSources.push_back(address);
}

void ZoneList::RemoveSource(const InetAddress & address)
{
	for(Sources::iterator i = iSources.begin(); i != iSources.end(); ++i)
		if(*i == address)
		{
			SSConnection * c = iPool.GetConnection(*i);
			if(c != 0)
				c->Disconnect();
			iSources.erase(i);
			break;
		}
}

void ZoneList::ServerUpdate(InetAddress & source, const BString & packet)
{
	try
	{
		Zone zone;
		zone.iIsMaster = true;
		zone.iIsUpdating = false;

		DataReader <LittleEndian, Ascii> reader(packet);
		reader.ReadUint32();
		zone.iAddress = InetAddress(source.AsUint32HostOrder(), reader.ReadUint16());
		zone.iPlayerCount = reader.ReadUint16();
		zone.iIsBilled = (reader.ReadUint16() != 0);
		zone.iServerVersion = reader.ReadUint32();
		zone.iName = reader.ReadString(32);
		zone.iPassword = reader.ReadString(48);

		for(;;)
		{
			int8 c = reader.ReadInt8();
			if(c == 0)
				break;
			zone.iDescription += c;
		}

		String::Trim(zone.iName);
		String::Trim(zone.iPassword);
		String::Trim(zone.iDescription);

		UpdateMasterZone(zone);

		Logger::Instance().Log(KLogInfo, "Received master update for zone '%s'.", zone.iName.c_str());
	}
	catch(const invalid_argument &)
	{
		Logger::Instance().Log(KLogWarning, "Invalid master update from %s.", source.AsString().c_str());
	}
}

void ZoneList::ClientRequest(SSConnection & connection, const BString & packet)
{
	if(packet.GetLength() != 5 || packet[0] != 0x01)
	{
		Logger::Instance().Log(KLogWarning, "Invalid zone list request from %s.", connection.GetRemoteAddress().AsString().c_str());
		connection.Disconnect();
		return;
	}

	DataReader <LittleEndian> reader(packet);
	reader.ReadUint8();
	uint32 minPop = reader.ReadUint32();

	Logger::Instance().Log(KLogDebug, "%s requesting zone list with min population of %d.", connection.GetRemoteAddress().AsString().c_str(), minPop);

	DataWriter <LittleEndian> writer;
	writer.AppendUint8(0x01);
	for(Zones::iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(!i->iIsDisabled && i->iPlayerCount >= minPop)
		{
			writer.AppendUint32(i->iAddress.AsUint32NetworkOrder());
			writer.AppendUint16(i->iAddress.GetPort());
			writer.AppendUint16(i->iPlayerCount);
			writer.AppendUint16(i->iIsBilled ? 0x01 : 0x00);
			writer.AppendUint32(i->iServerVersion);

			for(uint32 j = 0; j < 64; ++j)
				if(j < i->iName.length())
					writer.AppendInt8(i->iName[j]);
				else
					writer.AppendInt8(0x00);

			for(uint32 j = 0; j < i->iDescription.length(); ++j)
				writer.AppendInt8(i->iDescription[j]);

			writer.AppendInt8(0x00);
		}

	connection.SendFile(writer.AsBString());
}

void ZoneList::UpdateMasterZone(const Zone & zone)
{
	if(iConfig.IsZoneIpBanned(zone.iAddress) || iConfig.IsZoneNameBanned(zone.iName))
		return;

	uint32 now = Time::GetMilliCount();

	for(Zones::iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(i->iName == zone.iName)
		{
			//
			// Zones must be able to match the password or wait 24 hours
			//
			if(i->iIsMaster &&
			  (i->iPassword == zone.iPassword || (now - i->iUpdateTime) / 1000 >= iConfig.GetPasswordReclaimTime()))
			{
				uint32 min = std::min(i->iMinPlayerCount, zone.iPlayerCount);
				uint32 max = std::max(i->iMaxPlayerCount, zone.iPlayerCount);

				*i = zone;
				i->iIsMaster = true;
				i->iIsUpdating = false;
				i->iIsDisabled = false;
				i->iUpdateTime = now;
				i->iMinPlayerCount = min;
				i->iMaxPlayerCount = max;
			}
			return;
		}

	iZones.push_back(zone);
	Zone & z = iZones.back();

	z.iIsMaster = true;
	z.iIsUpdating = false;
	z.iIsDisabled = false;
	z.iUpdateTime = now;
	z.iMinPlayerCount = z.iPlayerCount;
	z.iMaxPlayerCount = z.iPlayerCount;
}

void ZoneList::Poll()
{
	iPool.Poll();

	//
	// Check for a zone list reply.
	//
	for(Sources::iterator i = iSources.begin(); i != iSources.end(); ++i)
	{
		SSConnection * c = iPool.GetConnection(*i);
		if(c == 0 || !c->HasMorePackets())
			continue;

		BString packet = c->GetNextPacket();
		if(packet.empty() || packet[0] != 0x01)
			continue;

		ProcessZoneListPacket(c->GetName(), packet);
		c->Disconnect();
	}

	DisableMissingMasters(Time::GetMilliCount());

	if(iZoneQueryTimer.GetTimeElapsed() / 1000 >= iConfig.GetZoneListRefreshPeriod())
	{
		ClearMirroredZones();
		UpdateZoneLists();

		iZoneQueryTimer.Reset();
	}
}

void ZoneList::DisableMissingMasters(uint32 currentTime)
{
	for(Zones::iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(i->iIsMaster && !i->iIsDisabled)
		{
			i->iIsDisabled = ((currentTime - i->iUpdateTime) > 5 * 60 * 1000);
			if(i->iIsDisabled)
				iLogger.Log(KLogInfo, "Purged zone '%s' due to inactivity.", i->iName.c_str());
		}
}

void ZoneList::UpdateMirroredZone(const Zone & zone)
{
	Zone * target = 0;
	for(Zones::iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(i->iName == zone.iName)
			target = &*i;

	uint32 min = 0, max = 0;
	if(target == 0)
	{
		iZones.push_back(zone);
		target = &iZones.back();
		min = target->iPlayerCount;
		max = target->iPlayerCount;
	}
	else if(target->iIsMaster || target->iIsUpdating)      // Update of zone in the sweep
	{
		min = std::min(target->iMinPlayerCount, zone.iPlayerCount);
		max = std::max(target->iMaxPlayerCount, zone.iPlayerCount);
		*target = zone;
	}
	else                                                   // Duplicate update in same sweep
	{
		target->iPlayerCount = std::max(target->iPlayerCount, zone.iPlayerCount);
		min = std::min(target->iMinPlayerCount, target->iPlayerCount);
		max = std::max(target->iMaxPlayerCount, target->iPlayerCount);
	}

	target->iIsMaster = false;
	target->iIsDisabled = false;
	target->iIsUpdating = false;
	target->iUpdateTime = Time::GetMilliCount();
	target->iMinPlayerCount = min;
	target->iMaxPlayerCount = max;
}

void ZoneList::ClearMirroredZones()
{
	//
	// Remove the zones that were marked as "updating" on the previous sweep
	// but no longer have entries.
	//
	for(Zones::iterator i = iZones.begin(); i != iZones.end(); )
	{
		Zones::iterator tmp = i++;
		if(!tmp->iIsMaster && tmp->iIsUpdating)
			iZones.erase(tmp);
	}

	for(Zones::iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(!i->iIsMaster)
			i->iIsUpdating = true;
}

void ZoneList::UpdateZoneLists()
{
	iLogger.Log(KLogDebug, "Updating zone list...");

	//
	// Drop all old connections.
	//
	for(Sources::iterator i = iSources.begin(); i != iSources.end(); ++i)
	{
		SSConnection * c = iPool.GetConnection(*i);
		if(c != 0)
		{
			iLogger.Log(KLogDebug, "Dropping stale connection to %s.", c->GetName().c_str());
			c->Disconnect();
		}
	}

	iPool.Poll();

	//
	// Make new connections and send out a request.
	//
	for(Sources::iterator i = iSources.begin(); i != iSources.end(); ++i)
	{
		SSConnection & c = iPool.Connect(i->AsString(), *i);
		c.Send(BString("\x01\x00\x00\x00\x00", 5), true);
	}
}

void ZoneList::ProcessZoneListPacket(const std::string & connName, const BString & packet)
{
	DataReader <LittleEndian, Ascii> reader(packet);
	reader.ReadUint8();

	uint32 count = 0;
	uint32 totalPop = 0;
	try
	{
		while(reader.BytesRemaining())
		{
			Zone zoneInfo;
			zoneInfo.iIsMaster = false;
			zoneInfo.iIsDisabled = false;
			zoneInfo.iIsUpdating = false;

			uint32 ipAddress = reader.ReadUint8() << 24;
			ipAddress |= reader.ReadUint8() << 16;
			ipAddress |= reader.ReadUint8() << 8;
			ipAddress |= reader.ReadUint8() << 0;

			zoneInfo.iAddress = InetAddress(ipAddress, reader.ReadUint16());
			zoneInfo.iPlayerCount = reader.ReadUint16();
			zoneInfo.iIsBilled = (reader.ReadUint16() != 0);
			zoneInfo.iServerVersion = reader.ReadUint32();
			zoneInfo.iName = reader.ReadString(64);

			for(;;)
			{
				int8 c = reader.ReadInt8();
				if(!c)
					break;
				zoneInfo.iDescription += c;
			}

			if(!iConfig.IsZoneIpBanned(zoneInfo.iAddress) && !iConfig.IsZoneNameBanned(zoneInfo.iName))
			{
				UpdateMirroredZone(zoneInfo);

				totalPop += zoneInfo.iPlayerCount;
				++count;
			}
		}
		iLogger.Log(KLogInfo, "%s delivered %d zones with a total population of %d.", connName.c_str(), count, totalPop);
	}
	catch(const invalid_argument &)
	{
		iLogger.Log(KLogWarning, "Invalid response from %s after %d zones parsed.", connName.c_str(), count);
	}
}

void ZoneList::WriteStatistics(const Time & serverLaunchTime)
{
	Logger::Instance().Log(KLogDebug, "Writing zone statistics.");

	try
	{
		File fp;
		fp.Open("stats/current.xml", "wt");

		std::stringstream content;
		content << "<DirectoryServer>"
		        << "<startTime>"
		        << serverLaunchTime.ToLocalTzString()
		        << "</startTime>";

		for(Zones::const_iterator i = iZones.begin(); i != iZones.end(); ++i)
		{
			if(i->iIsDisabled)
				continue;

			content << "<zone>"
			        << "<name>"
			        << UrlEncoding::DeserializeString(BString(i->iName))
			        << "</name>"
			        << "<description>"
			        << UrlEncoding::DeserializeString(BString(i->iDescription))
			        << "</description>"
			        << "<population>"
			        << i->iPlayerCount
			        << "</population>"
			        << "<maxPopulation>"
			        << i->iMaxPlayerCount
			        << "</maxPopulation>"
			        << "<minPopulation>"
			        << i->iMinPlayerCount
			        << "</minPopulation>"
			        << "</zone>";
		}

		content << "</DirectoryServer>";

		fp.Write(BString(content.str()));
	}
	catch(const sydus_error & error)
	{
		Logger::Instance().Log(KLogError, "Error writing stats/current.xml: %s", error.what());
	}

	Logger::Instance().Log(KLogDebug, "Writing population distribution statistics.");

	typedef std::vector <uint32> PopHistory;
	static PopHistory history;

	while(history.size() >= 24 * 60 / 1)
		history.erase(history.begin());

	uint32 population = 0;
	for(Zones::const_iterator i = iZones.begin(); i != iZones.end(); ++i)
		if(!i->iIsDisabled)
			population += i->iPlayerCount;
	history.push_back(population);

	try
	{
		File fp;
		fp.Open("stats/distribution.xml", "wt");

		std::stringstream content;
		content << "<DirectoryServer>"
		        << "<startTime>"
		        << serverLaunchTime.ToLocalTzString()
		        << "</startTime>"
		        << "<distribution>";

		for(PopHistory::iterator i = history.begin(); i != history.end(); ++i)
		{
			content << "<sample>"
			        << *i
			        << "</sample>";
		}

		content << "</distribution>";
		content << "</DirectoryServer>";

		fp.Write(BString(content.str()));
	}
	catch(const sydus_error & error)
	{
		Logger::Instance().Log(KLogError, "Error writing stats/distribution.xml: %s", error.what());
	}

	Logger::Instance().Log(KLogDebug, "Writing traditional statistics.");

	try
	{
		File fp;
		fp.Open("stats/stats.txt", "wt");

		std::stringstream content;
		for(Zones::const_iterator i = iZones.begin(); i != iZones.end(); ++i)
		{
			content << i->iAddress.GetIpAddress() << "\t"
			        << i->iAddress.GetPort() << "\t"
			        << i->iPlayerCount << "\t"
			        << (i->iIsBilled ? 1 : 0) << "\t"
			        << i->iName << "\t"
			        << i->iDescription << std::endl;
		}

		fp.Write(BString(content.str()));
	}
	catch(const sydus_error & error)
	{
		Logger::Instance().Log(KLogError, "Error writing stats/stats.txt: %s", error.what());
	}
}
